// C
#include <math.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

// C++
#include <typeinfo>
#include <sstream>
#include <string>
#include <vector>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <cstdlib>
#include <map>
#include <tuple>

// Boost
#include <boost/asio.hpp>
#include <boost/array.hpp>
#include <boost/bind.hpp>
#include <boost/thread/thread.hpp>
#include <boost/system/error_code.hpp>
#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/filesystem.hpp>
#include <boost/variant/recursive_variant.hpp>
#include <boost/foreach.hpp>
#include <boost/lambda/lambda.hpp>

// Spirit
#include <boost/config/warning_disable.hpp>
#include <boost/spirit/include/lex.hpp>
#include <boost/spirit/include/qi.hpp>
#include <boost/spirit/include/karma.hpp>
#include <boost/spirit/include/phoenix.hpp>

// Fusion
#include <boost/fusion/adapted/struct/adapt_struct.hpp>
#include <boost/fusion/include/adapt_struct.hpp>
#include <boost/fusion/include/io.hpp>

namespace test {
    struct item {
        int a;
        int b;
        int c;
    };

    typedef std::vector<item> items;

    void parse_items(items&);
}

BOOST_FUSION_ADAPT_STRUCT(
        test::item,
        (int, a)
        (int, b)
        (int, c)
        )	

namespace test {

        namespace fusion = boost::fusion;
        namespace phoenix = boost::phoenix;
        namespace qi = boost::spirit::qi;
        namespace ascii = boost::spirit::ascii;
        namespace lambda = boost::lambda;

        template <typename InputIterator>
            struct skipper_grammar
            : qi::grammar<InputIterator>
            {
                typedef skipper_grammar<InputIterator> type;

                skipper_grammar()
                    : skipper_grammar::base_type( skip_on ),
                    comment_char( '#' )
                {
                    using ascii::char_;
                    using ascii::space;
                    using ascii::no_case;
                    using qi::lit;
                    using qi::eol;
                    using phoenix::ref;
                    using namespace qi::labels;

                    header = lit("\"recType\"")
                        | lit("\"comment\"")
                        | lit(",,,,,,,,,,")
                        ;

                    comment =
                        (
                         char_( ref( comment_char ) )
                         | header
                         >> *( char_ - eol ) >> eol
                        )
                        | lit("\"data\"")
                        ;

                    comment_charset =
                        char_( "!\"#$%&'()*,:;<>?@\\^`{|}~" )
                        ;

                    skip_on =
                        space  // tab/space/cr/lf
                        | comment
                        ;
                }

                char comment_char;
                qi::rule<InputIterator> comment;
                qi::rule<InputIterator, char()> comment_charset;
                qi::rule<InputIterator, void()> header;
                qi::rule<InputIterator> comment_directive;
                qi::rule<InputIterator> skip_on;
            };


        template<typename Iterator, typename Skipper>
            struct item_parser : qi::grammar<Iterator, items(), Skipper>
        {
            item_parser() : item_parser::base_type(a ,"a")
            {
                using qi::lit;
                using qi::repeat;
                using qi::inf;
                using qi::on_error;
                using qi::eol;
                using qi::fail;
                using qi::int_;
                using namespace qi::labels;
                using phoenix::construct;
                using phoenix::val;
                using phoenix::if_;
                using phoenix::arg_names::arg1;

                b %= '[' > int_ > ']' > '[' > int_ > ']' > '<' > int_ > '>';

                a %= *b;

                b.name("item");
                a.name("items");

                on_error<fail>
                    (
                     a,
                     std::cout
                     << val("Error! Expecting ")
                     << boost::spirit::_4   // what failed
                     << val(" here: \"")
                     << construct<std::string>(boost::spirit::_3, boost::spirit::_2) // interators to error pos, end
                     << val("\"")
                     << std::endl
                    );

            }

            qi::rule<Iterator, items(), Skipper> a;
            qi::rule<Iterator, item(), Skipper> b;
        }; 

    }

void test::parse_items(items &_items)
{
    // LOAD FILE
    std::ifstream fs;
    fs.open("test.txt", std::ios::in);

    if(!fs)
    {
        std::cout << "Error! could not load file";
    }

    using boost::spirit::ascii::space;
    typedef std::string::const_iterator iterator_type;
    typedef skipper_grammar<iterator_type> skipper_type;
    typedef item_parser<iterator_type, skipper_type> item_parser;

    item_parser g; // our grammar

    long length;
    char* buffer;

    // get length of file
    fs.seekg(0, std::ios::end);
    length = fs.tellg();
    fs.seekg(0, std::ios::beg);

    // allocate memory
    buffer = new char[length+1];

    // read data
    fs.read(buffer, length);
    buffer[length]='\0';
    fs.close();

    std::string input = buffer;
    delete [] buffer;

    std::string::const_iterator iter = input.begin();
    std::string::const_iterator end = input.end();
    bool r = phrase_parse(iter, end, g, skipper_type(), _items);

    if(r && iter == end)
    { 
        //std::cout << "alarm log succeeded parse\n";
    }
    else
    {
        std::cout << "failed to parse\n";
    }
}


int main (int argc, char *argv[])
{
    test::items i;
    parse_items(i);
    return 0;
}

// COMMITING OUT THIS BOTTOM LINE WILL GET
// RID OF BOOST FUSION VECTOR ERROR
#include "main.moc"
